import numpy as np
import shap

from tsproto.windowshap import SlidingWindowSHAP

"""
Source code taken from: https://github.com/vsubbian/WindowSHAP
"""
def outliers(data, multiplier=1.5):
    """

    :param data:
    :param multiplier:
    :return:
    """
    # finding the 1st quartile
    q1 = np.quantile(data, 0.25)

    # finding the 3rd quartile
    q3 = np.quantile(data, 0.75)

    # finding the iqr region
    iqr = q3 - q1

    # finding upper and lower whiskers
    upper_bound = q3 + (multiplier * iqr)
    lower_bound = q1 - (multiplier * iqr)
    return (iqr, upper_bound, lower_bound)

def dominant_frequency(time_series_row, sampling_rate):
    """
    Calculate the dominant frequency of a time series row using FFT.

    Parameters:
    - time_series_row: 1D numpy array representing the time series data.
    - sampling_rate: The sampling rate of the time series.

    Returns:
    - dominant_freq: Dominant frequency of the time series row.
    """
    # Check for NaN values and replace them with zeros
    time_series_row = np.nan_to_num(time_series_row)

    # Compute the FFT
    fft_result = np.fft.fft(time_series_row)

    # Calculate the corresponding frequencies
    frequencies = np.fft.fftfreq(len(time_series_row), d=1 / sampling_rate)

    # Find the index of the maximum amplitude in the positive frequencies
    positive_freq_mask = frequencies > 0
    dominant_freq_index = np.argmax(np.abs(fft_result[positive_freq_mask]))

    # Get the dominant frequency
    dominant_freq = frequencies[positive_freq_mask][dominant_freq_index]

    return np.abs(dominant_freq)


def dominant_frequencies_for_rows(time_series_array, sampling_rate):
    """
    Calculate the dominant frequency for each row of a 2D numpy array.

    Parameters:
    - time_series_array: 2D numpy array where each row represents a time series.
    - sampling_rate: The sampling rate of the time series.

    Returns:
    - dominant_frequencies: 1D numpy array containing the dominant frequency for each row.
    """
    # Apply the dominant_frequency function to each row
    dominant_frequencies = np.apply_along_axis(dominant_frequency, axis=1, arr=time_series_array,
                                               sampling_rate=sampling_rate)

    return dominant_frequencies



def getshap(model, X, y, shap_version='deep', bg_size=1000, stride=10, window_len=10, absshap=True, shuffle=True):
    indexes = np.arange(0, len(X))
    if shuffle:
        np.random.shuffle(indexes)
    maxid = min(bg_size, len(X))

    background_data = X[indexes[:maxid]]
    if shap_version == 'window':
        sv_tr = np.zeros((len(X), X.shape[1], X.shape[2]))

        for i in range(len(X)):
            gtw = SlidingWindowSHAP(model, stride, window_len, background_data, X[i:i + 1], model_type='lstm')
            sv_tr[i, :, :] = gtw.shap_values(num_output=y.shape[1])
    elif shap_version == 'deep':
        explainer = shap.DeepExplainer(model, background_data)
        shap_values_tr = explainer.shap_values(X, check_additivity=False)
        if absshap:
            sv_tr = abs(np.array(shap_values_tr)).mean(
                axis=0)  # This basically returns the average importance over the feature/sample
            # Not taking into account the sign of shap value, as it is not required
            # for breakpoints calculation
        else:
            indexer = np.argmax(model.predict(X), axis=1)
            sv_tr = []
            for i in range(0, len(X)):
                sv_tr.append([shap_values_tr[indexer[i]][i, :]])
            sv_tr = np.concatenate(sv_tr)
        return background_data,sv_tr


import warnings
import ruptures as rpt
from tslearn.clustering import TimeSeriesKMeans
from tslearn.utils import to_time_series_dataset
from sklearn.tree import DecisionTreeClassifier
import time
from tslearn.clustering import KShape
from kshape.core import KShapeClusteringCPU
from tslearn.preprocessing import TimeSeriesScalerMeanVariance
from sklearn.cluster import KMeans
from sklearn.metrics import f1_score, precision_score, recall_score, accuracy_score
from sklearn.base import BaseEstimator, TransformerMixin
import numpy as np
import pandas as pd
from tsproto.utils import dominant_frequencies_for_rows

"""
Documentation of this module

"""


class PrototypeEncoder(BaseEstimator, TransformerMixin):
    """
    Encodes time-series into prototypes
    """

    def __init__(self, blackbox, min_size, jump, pen, n_clusters, multiplier=1.5, method='dtw',
                 descriptors=['existance', 'duration', 'stats', 'startpoint'], n_jobs=None, verbose=0,
                 dims=1, sampling_rate=1, feature_names=None, importance_aggregation_func=np.mean):
        """ Initializes PrototypeEncoder class

        :param blackbox: instance of a blackbox model that is to be explained
        :param min_size: minimum size of a prototype (Pelt algorithm parameter)
        :param jump: subsample (one every jump points) (Pelt algorithm parameter)
        :param pen: penalty value (>0) (Pelt algorithm parameter)
        :param n_clusters: number of clusters to generated (these are going to be prototypes). It can be int, float or dict.
        :param multiplier: multiplier used in outlier deteciton. The smaller the value the stronger reduction of outliers
        :param method: clustering algorithms method. Default dtw.
        :param descriptors: what description functions use to describe prototypes.
        :param n_jobs: parallelization. Default None
        :param verbose: verbosity level. Default 0
        :param dims: number of dimensions/features in time series. Default 1
        :param sampling_rate: sampling rate used to calculate dominant frequency. Default 1
        :param feature_names: list of feature names. Default None
        :param importance_aggregation_func: function to aggregate shap values. Default np.mean
        """
        self.threshold = 0  # threshold for discarding slices of time-series of low importance
        self.multiplier = multiplier
        self.n_clusters = {}
        self.jump = jump
        self.min_size = min_size
        self.method = method
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.pen = pen
        self.dims = dims
        self.blackbox = blackbox
        self.descriptors = descriptors
        self.thresholds_ = [None] * dims
        self.kms_ = [None] * dims
        self.label_features_ = [None] * dims
        self.tsms_ = [None] * dims
        self.xbis_ = {}
        self.xbis_shap_ = {}
        self.xbis_cluster_labels_ = {}
        self.signal_ids_ = {}
        self.sampling_rate = sampling_rate
        self.weights_ = {}
        self.seq_len_ = {}
        self.importance_aggregation_func = importance_aggregation_func

        if self.method == 'kshapegpu':
            try:
                from kshape.core_gpu import KShapeClusteringGPU
            except ImportError:
                self.method = 'kshape'
                warnings.warn(
                    "To use GPU version of KSHAPE, install TSProto with GPU support: pip install tsproto[gpu]")

        if feature_names is None:
            self.feature_names = [str(f) for f in range(0, dims)]
        else:
            self.feature_names = feature_names

        # Make the number of clusters be determined by the number of breakpoints in data
        if n_clusters is None:
            n_clusters = 1

        # If the n_clusters was given but it is not dict, assign same number of clusters to all
        # dmiensions. It can be a fraction, which will make the algorithm determine the optimal number of clusters
        # while fitting. The fraction represetns the desired granularity of clustersng.
        # Otherwise, just assigne the dict to self.n_clusters this allows for a mixture of
        # fixed number of clusters and fractions (automatic detection)
        if not isinstance(n_clusters, dict):
            for dim in range(dims):
                self.n_clusters[self.feature_names[dim]] = n_clusters
        else:
            self.n_clusters = n_clusters

    def fit(self, X, y=None, shapclass=None):
        """ Fits model to the training data

        :param X:
        :type X: array-like, shape (n_samples, n_timestamps, n_features)
        :param y: not used, kept only for the compliance with scikit-learn API
        :type y: array-like
        :param shapclass: SHAP value calculated for the blackbox model to be explained.
        It is assumed that the mean absolute values are passed here.
        :type shapclass: array-like, shape (n_samples,n_timestamps,)
        :return:
        """
        if y is not None and shapclass is None:
            shapclass = y
        return self._transform(X, shapclass, refit=True, transform=False)

    def transform(self, X, y=None, shapclass=None):
        """

        :param X:
        :param y:
        :param shapclass:
        :return:
        """

        if y is not None and shapclass is None:
            shapclass = y
        return self._transform(X, shapclass, refit=False, transform=True)

    def fit_transform(self, X, y=None, shapclass=None):
        """

        :param X:
        :param y:
        :param shapclass:
        :return:
        """
        if y is not None and shapclass is None:
            shapclass = y
        return self._transform(X, shapclass, refit=True, transform=True)

    def _transform(self, Xdim, shap_dim, refit=False, transform=True):
        """

        :param Xdim:
        :param shap_dim:
        :param refit:
        :param transform:
        :return:
        """
        train_dim = []
        features = []
        if self.verbose > 0:
            print(f'Dataset shape: {Xdim.shape}')
            start_time = time.time()
            print(f'Calculating changepoints...')

        for dim in range(0, Xdim.shape[2]):
            totalslice = []
            totalsslice = []
            totalbpoints = []
            indexes = []
            print(f'Dim: {dim}')
            X = Xdim[:, :, dim].reshape((Xdim.shape[0], Xdim.shape[1], 1))
            shapclass = shap_dim[:, :, dim].reshape((shap_dim.shape[0], shap_dim.shape[1], 1))
            sum_bkps = 0
            for i in range(0, X.shape[0]):
                algo = rpt.Pelt(model="rbf", min_size=self.min_size, jump=self.jump).fit(shapclass[i])

                try:
                    breakpoints = algo.predict(pen=1)
                except:
                    # no change detected, means there are no breakpoints
                    breakpoints = [0]

                sum_bkps += len(breakpoints)
                shapslices = [abs(s) for s in np.split(shapclass[i], breakpoints) if
                              len(s) != 0]  # (assuming this is abshap over all classses)
                slices = [s for s in np.split(X[i], breakpoints) if len(s) != 0]
                bpoints = list(set([0] + [s for s in breakpoints if s < len(X[i]) and s > 0]))

                totalsslice.append(shapslices)
                totalslice.append(slices)
                totalbpoints.append(bpoints)
                indexes.append(np.ones(len(slices)) * i)

            print(len(totalsslice[0]))
            print(len(totalslice[0]))
            print(len(indexes[0]))
            print(len(totalbpoints[0]))

            ############ overriding clusters\
            if refit and (self.n_clusters[self.feature_names[dim]] <= 1):
                self.n_clusters[self.feature_names[dim]] = max(2, int(int(sum_bkps / X.shape[0]) * self.n_clusters[
                    self.feature_names[dim]]))
                print(f'For {self.feature_names[dim]} c_clusters = {self.n_clusters[self.feature_names[dim]]}')

            ############# .
            if refit:
                if self.multiplier is not None:
                    mean_concat_sslices = np.array(
                        [self.importance_aggregation_func(item) for sublist in totalsslice for item in sublist])
                    self.thresholds_[dim] = np.mean(mean_concat_sslices) - self.multiplier * np.std(mean_concat_sslices)
                else:
                    self.thresholds_[dim] = 0

            # ante-hoc filtering
            print(f'Before filtering no sigids: {(len(np.unique(np.concatenate(indexes))))}, uniqu {0}')
            if self.thresholds_[dim] > 0:
                indexed_slices = [[i, s, ss, bp] for ts, tss, ti, bpi in zip(totalslice, totalsslice, indexes, totalbpoints)
                                  for s, ss, i, bp in zip(ts, tss, ti, bpi) if
                                  self.importance_aggregation_func(abs(ss)) >= self.thresholds_[dim]]
                isdf = pd.DataFrame(indexed_slices)
                indexes_new = isdf.loc[:, 0]
                print(
                    f'Unique indexes: {len(np.unique(indexes_new))} vs max+1: {(len(np.unique(np.concatenate(indexes))))}')
                if len(np.unique(indexes_new)) < (len(np.unique(np.concatenate(indexes)))):
                    # in case there is a signal that has absolutely no readings
                    # reduce the threshold
                    print('WARNING: Changing the threshold, due to empty record')
                    full_slices = [[i, s, ss, bp] for ts, tss, ti, bpi in zip(totalslice, totalsslice, indexes, totalbpoints)
                                   for s, ss, i, bp in zip(ts, tss, ti, bpi)]
                    isdf = pd.DataFrame(full_slices, columns=['index', 'slice', 'shapslice', 'breakpoint'])
                    isdf['maxshap'] = isdf['shapslice'].apply(lambda x: np.mean(abs(x)))
                    self.thresholds_[dim] = isdf.groupby('index')['maxshap'].max().min()
                    indexed_slices = [[i, s, ss, bp] for ts, tss, ti, bpi in
                                      zip(totalslice, totalsslice, indexes, totalbpoints) for s, ss, i, bp in
                                      zip(ts, tss, ti, bpi) if
                                      self.importance_aggregation_func(abs(ss)) >= self.thresholds_[dim]]
                    print(f'Threshold: {self.thresholds_[dim]}')
                    isdf = pd.DataFrame(indexed_slices)
                indexes = list(isdf.groupby(0)[0].apply(np.array))
                totalslice = list(isdf.groupby(0)[1].apply(list))
                totalsslice = list(isdf.groupby(0)[2].apply(list))
                totalbpoints = list(isdf.groupby(0)[3].apply(np.array))

            ############# post-hoc filtering
            # TODO: filter whole clusters which average/max importance is below certain point
            print(f'After removing , no sigids: {(len(np.unique(np.concatenate(indexes))))}, uniqu {0}')

            totalslice = [item for sublist in totalslice for item in sublist]
            totalsslice = [item for sublist in totalsslice for item in
                           sublist]  # this can be treated as weights in samples

            if self.verbose > 0:
                end_time = time.time()
                print(f'Done in {(end_time - start_time)}.')
                start_time = time.time()
                print(f'Clustering data')
            X_bis_o = to_time_series_dataset(totalslice)
            X_bis_o_shap = to_time_series_dataset(totalsslice)

            print(X_bis_o.shape)
            print(X_bis_o_shap.shape)

            if not refit:
                if self.method != 'dtw':
                    cur_seq_len = X_bis_o.shape[1]
                    if cur_seq_len < self.seq_len_[dim]:
                        diff = self.seq_len_[dim] - cur_seq_len
                        if len(X_bis_o.shape) == 3:
                            expand = np.zeros((X_bis_o.shape[0], diff, X_bis_o.shape[2]))
                        elif len(X_bis.shape) == 2:
                            expand = np.zeros((X_bis_o.shape[0], diff))
                        X_bis = np.append(X_bis_o, expand, axis=1)
                        X_bis_shap = np.append(X_bis_o_shap, expand, axis=1)
                    elif cur_seq_len > self.seq_len_[dim]:
                        diff = cur_seq_len - self.seq_len_[dim]
                        X_bis = np.delete(X_bis_o, slice(X_bis_o.shape[1] - diff, None), 1)
                        X_bis_shap = np.delete(X_bis_o_shap, slice(X_bis_o_shap.shape[1] - diff, None), 1)
                    else:
                        X_bis = X_bis_o
                        X_bis_shap = X_bis_o_shap
                else:
                    X_bis = X_bis_o
                    X_bis_shap = X_bis_o_shap
            else:
                X_bis = X_bis_o
                X_bis_shap = X_bis_o_shap

            if self.method in ['tskshape', 'kmeans']:
                X_bis = np.nan_to_num(X_bis)
            if self.method in ['kshape', 'tskshape', 'kshapegpu', 'kmeans']:
                if refit:
                    self.tsms_[dim] = TimeSeriesScalerMeanVariance()
                    X_bis = self.tsms_[dim].fit_transform(X_bis)
                else:
                    X_bis = self.tsms_[dim].transform(np.nan_to_num(X_bis))
            if self.method == 'kmeans':
                X_bis = X_bis.reshape(X_bis.shape[0], X_bis.shape[1])
                X_bis_shap = X_bis_shap.reshape(X_bis_shap.shape[0], X_bis_shap.shape[1])

            if self.verbose > 0:
                print(f'Shape of data for clustering: {X_bis.shape}')

            if refit:
                if self.verbose:
                    start_time = time.time()
                    print(f'Clustering data')
                if self.method == 'dtw':
                    self.kms_[dim] = TimeSeriesKMeans(n_clusters=self.n_clusters[self.feature_names[dim]], metric="dtw",
                                                      max_iter=5,
                                                      random_state=0, n_jobs=self.n_jobs)
                elif self.method == 'tskshape':
                    self.kms_[dim] = KShape(n_clusters=self.n_clusters[self.feature_names[dim]], verbose=True)
                elif self.method == 'kshape':
                    self.kms_[dim] = KShapeClusteringCPU(n_clusters=self.n_clusters[self.feature_names[dim]],
                                                         n_jobs=self.n_jobs)
                elif self.method == 'kshapegpu':
                    from kshape.core_gpu import KShapeClusteringGPU
                    self.kms_[dim] = KShapeClusteringGPU(n_clusters=self.n_clusters[self.feature_names[dim]])
                elif self.method == 'kmeans':
                    self.kms_[dim] = KMeans(n_clusters=self.n_clusters[self.feature_names[dim]])

                print(f'X_bis shape: {X_bis.shape} for method={self.method}')
                self.seq_len_[dim]=X_bis.shape[1]

                self.kms_[dim].fit(X_bis)

                if self.method == 'kshape':
                    self.kms_[dim].cluster_centers_ = self.kms_[dim].centroids_
                if self.method == 'kshapegpu':
                    self.kms_[dim].cluster_centers_ = self.kms_[dim].centroids_.detach().cpu()

                self.xbis_[self.feature_names[dim]] = X_bis
                self.xbis_shap_[self.feature_names[dim]] = X_bis_shap
                self.xbis_cluster_labels_[self.feature_names[dim]] = self.kms_[dim].labels_

            if self.verbose > 0:
                end_time = time.time()
                print(f'Done in {(end_time - start_time)}.')
                start_time = time.time()
                print(f'OHE time series')
            signal = np.concatenate(indexes)
            Xdf = pd.DataFrame(signal, columns=['sigid'])

            if refit:
                labels = self.kms_[dim].labels_
                self.label_features_[dim] = np.arange(0, np.max(labels) + 1)
                self.signal_ids_[self.feature_names[dim]] = np.concatenate(indexes)
            else:
                labels = self.kms_[dim].predict(X_bis)
                # todo xbis and cluster centers should be changed here?

            if not transform:
                return self

            Xdf['cluster'] = labels
            Xdf['shapweight'] = np.array([self.importance_aggregation_func(abs(sv)) for sv in totalsslice])
            Xdf['startpoint'] = np.concatenate(totalbpoints)
            Xdf['durations'] = X_bis_o.shape[1] - np.sum(np.isnan(X_bis), axis=1)

            Xdf['min'] = np.nanmin(X_bis, axis=1)
            Xdf['max'] = np.nanmax(X_bis, axis=1)
            Xdf['mean'] = np.nanmean(X_bis, axis=1)
            Xdf['std'] = np.nanstd(X_bis, axis=1)
            Xdf['frequency'] = dominant_frequencies_for_rows(X_bis, sampling_rate=self.sampling_rate)

            phantom = pd.DataFrame({'sigid': [-1] * len(self.label_features_[dim]),
                                    'cluster': np.arange(0, len(self.label_features_[dim])),
                                    'startpoint': np.arange(0, len(self.label_features_[dim])),
                                    'durations': [0] * len(self.label_features_[dim]),
                                    'min': [0] * len(self.label_features_[dim]),
                                    'max': [0] * len(self.label_features_[dim]),
                                    'mean': [0] * len(self.label_features_[dim]),
                                    'std': [0] * len(self.label_features_[dim]),
                                    'frequency': [0] * len(self.label_features_[dim])
                                    })

            Xdfp = pd.concat((Xdf, phantom))
            # print(f'Labels: {labels} label_features: {self.label_features_[dim]} and labels from km {np.unique(self.kms_[dim].labels_)}')

            ohe_train = pd.pivot_table(Xdfp, index='sigid', columns='cluster', values='durations',
                                       aggfunc=lambda x: sum(~np.isnan(x))).fillna(0).astype(int)
            startpoint_train = pd.pivot_table(Xdfp, index='sigid', columns='cluster', values='startpoint',
                                              aggfunc='min').fillna(0).astype(int)
            duration_train = pd.pivot_table(Xdfp, index='sigid', values='durations', columns='cluster').fillna(
                0)  # TODO: wrong aggfunc?
            min_train = pd.pivot_table(Xdfp, index='sigid', values='min', columns='cluster').fillna(0)
            max_train = pd.pivot_table(Xdfp, index='sigid', values='max', columns='cluster').fillna(0)
            mean_train = pd.pivot_table(Xdfp, index='sigid', values='mean', columns='cluster').fillna(0)
            std_train = pd.pivot_table(Xdfp, index='sigid', values='std', columns='cluster').fillna(0)
            frequency_train = pd.pivot_table(Xdfp, index='sigid', values='frequency', columns='cluster').fillna(0)
            self.weights_[dim] = Xdf.groupby('sigid')['shapweight'].apply(self.importance_aggregation_func).values
            print(f'Shape weights: {self.weights_[dim].shape}')

            # print(f'Columns: {ohe_train.columns}')

            ohe_train.columns = [f'exists_cl_{c}_{self.feature_names[dim]}' for c in
                                 range(0, len(self.label_features_[dim]))]
            duration_train.columns = [f'duration_cl_{c}_{self.feature_names[dim]}' for c in
                                      range(0, len(self.label_features_[dim]))]
            min_train.columns = [f'min_cl_{c}_{self.feature_names[dim]}' for c in
                                 range(0, len(self.label_features_[dim]))]
            max_train.columns = [f'max_cl_{c}_{self.feature_names[dim]}' for c in
                                 range(0, len(self.label_features_[dim]))]
            mean_train.columns = [f'mean_cl_{c}_{self.feature_names[dim]}' for c in
                                  range(0, len(self.label_features_[dim]))]
            std_train.columns = [f'std_cl_{c}_{self.feature_names[dim]}' for c in
                                 range(0, len(self.label_features_[dim]))]
            frequency_train.columns = [f'frequency_cl_{c}_{self.feature_names[dim]}' for c in
                                       range(0, len(self.label_features_[dim]))]
            startpoint_train.columns = [f'startpoint_cl_{c}_{self.feature_names[dim]}' for c in
                                        range(0, len(self.label_features_[dim]))]

            train = pd.concat((ohe_train, duration_train, min_train, max_train, mean_train, std_train, frequency_train,
                               startpoint_train), axis=1)
            train = train[~train.index.isin([-1])]

            if 'existance' in self.descriptors:
                features = features + list(ohe_train.columns)
            if 'duration' in self.descriptors:
                features = features + list(duration_train.columns) + list(startpoint_train.columns)
            if 'stats' in self.descriptors:
                features = features + list(min_train.columns) + list(max_train.columns) + list(
                    mean_train.columns) + list(std_train.columns)
            if 'frequency' in self.descriptors:
                features = features + list(frequency_train.columns)

            train_dim.append(train)

        train = pd.concat(train_dim, axis=1).fillna(0)
        target = 'target'

        print(f'Len X={len(X)} vs len of train = {len(train)}')
        bbox_predictions = self.blackbox.predict(Xdim)
        if len(bbox_predictions.shape) == 2:
            train[target] = np.argmax(bbox_predictions, axis=1)
        else:
            train[target] = bbox_predictions
        if self.verbose > 0:
            end_time = time.time()
            print(f'Done in {(end_time - start_time)}.')
        weights = pd.DataFrame(pad_arrays_in_dict(self.weights_)).sum(axis=1).values
        return train, features, target, weights


def pad_arrays_in_dict(array_dict):
    # Find the maximum length among the arrays
    max_length = max(len(array) for array in array_dict.values())

    # Pad arrays with zeros to make them of equal length
    padded_arrays = {label: np.pad(array, (0, max_length - len(array)), mode='constant', constant_values=0)
                     for label, array in array_dict.items()}

    return padded_arrays


class InterpretableModel:
    def fit_or_predict(self, ohe_train, features, target, intclf=None, verbose=0, max_depth=None, min_samples_leaf=0.05,
                       weights=None, average=None):
        if intclf is None:
            intclf = DecisionTreeClassifier(max_depth=max_depth, min_samples_leaf=min_samples_leaf)
            intclf.fit(ohe_train[features], ohe_train[target], sample_weight=weights)
        iclf_pred = intclf.predict(ohe_train[features])
        if verbose > 0:
            print(f'Tree depth: {intclf.get_depth()}')
        if average is None and max(len(np.unique(iclf_pred)), ohe_train[target].nunique()) > 2:
            average = 'macro'
        elif average is None:
            average = 'binary'
        return (accuracy_score(ohe_train[target], iclf_pred),
                precision_score(ohe_train[target], iclf_pred, average=average),
                recall_score(ohe_train[target], iclf_pred, average=average),
                f1_score(ohe_train[target], iclf_pred, average=average),
                intclf)



import math
import numpy as np
import shap
from copy import deepcopy
import warnings


class StationaryWindowSHAP():
    '''
    A class for computing the shapely values for time sereis data. Only the shap values for the first output
    is reported.

    Parameters:
    model: A model object that will be used for prediction. The model object must have a method called predict() which produces the model output for a given input
    window_len: The length of the window for the algorithm
    B_ts: A 3D numpy array of background time series data
    test_ts: A 3D numpy array of test time series data
    B_mask: A 3D numpy array of background masking data. It is only used for specific models such as GRUD where a masking variable is passed to the model alongside the time series data. (default: None)
    B_dem: A 2D numpy array of background demographic data (non-temporal data). It is only used for specific models with both modelities of temporal and non-temporal variables. (default: None)
    test_mask: A 3D numpy array of test mask data (default: None)
    test_dem: A 2D numpy array of test demographic data (default: None)
    model_type: The type of model being used. Set the parameter to 'lstm' when time series data is the only input, pick 'lstm_dem' when input includes both time sereis and demographic (non-termporal) data, and 'grud' when you are using GRUD structure.  (default: 'lstm')
    '''

    def __init__(self, model, window_len, B_ts, test_ts, B_mask=None, B_dem=None,
                 test_mask=None, test_dem=None, model_type='lstm'):
        self.model = model
        self.window_len = window_len
        self.num_window = np.ceil(B_ts.shape[1] / self.window_len).astype('int')
        self.num_background = len(B_ts)
        self.num_test = len(test_ts)
        self.background_ts = B_ts
        self.background_mask = B_mask
        self.background_dem = B_dem
        self.test_ts = test_ts
        self.test_mask = test_mask
        self.test_dem = test_dem
        self.model_type = model_type
        self.ts_phi = None
        self.dem_phi = None
        self.explainer = None

        # Problem sizes
        self.num_ts_ftr = B_ts.shape[2]
        self.num_ts_step = B_ts.shape[1]
        self.num_dem_ftr = 0 if B_dem is None else B_dem.shape[1]

        # Creating all data (background and test together)
        self.all_ts = np.concatenate((self.background_ts, self.test_ts), axis=0)
        self.all_mask = None if test_mask is None else np.concatenate((self.background_mask, self.test_mask), axis=0)
        self.all_dem = None if test_dem is None else np.concatenate((self.background_dem, self.test_dem), axis=0)

        # Creating converted data for SHAP
        self.background_data = self.data_prepare(ts_x=self.background_ts, dem_x=self.background_dem, start_idx=0)
        self.test_data = self.data_prepare(ts_x=self.test_ts, dem_x=self.test_dem, start_idx=self.num_background)

    def data_prepare(self, ts_x, dem_x=None, start_idx=0):
        assert len(ts_x.shape) == 3
        assert dem_x is None or len(dem_x.shape) == 2
        dem_len = 0 if dem_x is None else dem_x.shape[1]

        total_num_features = self.num_dem_ftr + self.num_ts_ftr * self.num_window

        x_ = [[i] * total_num_features for i in range(start_idx, start_idx + ts_x.shape[0])]

        return np.array(x_)

    def wraper_predict(self, x):
        assert len(x.shape) == 2

        dem_x, ts_x = x[:, :self.num_dem_ftr].copy(), x[:, self.num_dem_ftr:].copy()

        # initializing the value of all arrays
        ts_x_ = np.zeros((x.shape[0], self.all_ts.shape[1], self.all_ts.shape[2]))
        mask_x_ = np.zeros_like(ts_x_)
        dem_x_ = np.zeros_like(dem_x, dtype=float)
        tstep = np.ones((x.shape[0], self.all_ts.shape[1], 1)) * \
                np.reshape(np.arange(0, self.all_ts.shape[1]), (1, self.all_ts.shape[1], 1))

        # Reshaping the ts indices based on the num time windows and features
        ts_x = ts_x.reshape((ts_x.shape[0], self.num_window, self.num_ts_ftr))

        for i in range(x.shape[0]):
            # creating time series data
            for t in range(self.num_ts_step):
                for j in range(self.num_ts_ftr):
                    # Finding the corresponding time interval
                    wind_t = np.ceil((t + 1) / self.window_len).astype('int') - 1
                    ind = ts_x[i, wind_t, j]
                    ts_x_[i, t, j] = self.all_ts[ind, t, j]
                    mask_x_[i, t, j] = None if self.all_mask is None else self.all_mask[ind, t, j]
            # creating static data
            for j in range(dem_x.shape[1]):
                ind = dem_x[i, j]
                dem_x_[i, j] = None if self.all_dem is None else self.all_dem[ind, j]

        # Creating the input of the model based on the different models.
        # This part should be updated as new models get involved in the project
        if self.model_type == 'lstm_dem':
            model_input = [ts_x_, dem_x_]
        elif self.model_type == 'grud':
            model_input = [ts_x_, mask_x_, tstep]
        elif self.model_type == 'lstm':
            model_input = ts_x_

        return self.model.predict(model_input)

    def shap_values(self, num_output=1):
        self.explainer = shap.KernelExplainer(self.wraper_predict, self.background_data)
        shap_values = self.explainer.shap_values(self.test_data)
        shap_values = np.array(shap_values)

        self.dem_phi = shap_values[:, :, :self.num_dem_ftr]
        ts_shap_values = shap_values[:, :, self.num_dem_ftr:]
        self.ts_phi = ts_shap_values.reshape((num_output, self.num_test, self.num_window, self.num_ts_ftr))

        # assign values to each single time step by deviding the values by window length
        self.ts_phi = np.repeat(self.ts_phi / self.window_len, self.window_len, axis=2)[:, :, :self.num_ts_step, :]

        # Reporting only the first output
        self.ts_phi = self.ts_phi[0]
        self.dem_phi = self.dem_phi[0]

        return self.ts_phi if self.num_dem_ftr == 0 else (self.dem_phi, self.ts_phi)


class SlidingWindowSHAP():
    '''
    A class for computing the shapely values for time sereis data. Only the shap values for the first output
    is reported.

    Parameters:
    model: A model object that will be used for prediction. The model object must have a method called predict() which produces the model output for a given input
    stride: The stride parameter for the Sliding WindowSHAP algorithm
    window_len: The length of the window for the algorithm
    B_ts: A 3D numpy array of background time series data
    test_ts: A 3D numpy array of test time series data
    B_mask: A 3D numpy array of background masking data. It is only used for specific models such as GRUD where a masking variable is passed to the model alongside the time series data. (default: None)
    B_dem: A 2D numpy array of background demographic data (non-temporal data). It is only used for specific models with both modelities of temporal and non-temporal variables. (default: None)
    test_mask: A 3D numpy array of test mask data (default: None)
    test_dem: A 2D numpy array of test demographic data (default: None)
    model_type: The type of model being used. Set the parameter to 'lstm' when time series data is the only input, pick 'lstm_dem' when input includes both time sereis and demographic (non-termporal) data, and 'grud' when you are using GRUD structure.  (default: 'lstm')
    '''

    def __init__(self, model, stride, window_len, B_ts, test_ts, B_mask=None,
                 B_dem=None, test_mask=None, test_dem=None, model_type='lstm'):
        self.model = model
        self.model_type = model_type
        self.stride = stride
        self.window_len = window_len
        self.num_window = 2  # Specific to the sliding time window
        self.num_background = len(B_ts)
        self.num_test = len(test_ts)
        self.background_ts = B_ts
        self.background_mask = B_mask
        self.background_dem = B_dem
        self.test_ts = test_ts
        self.test_mask = test_mask
        self.test_dem = test_dem
        self.ts_phi = None
        self.dem_phi = None
        self.explainer = None

        # Problem sizes
        self.num_ts_ftr = B_ts.shape[2]
        self.num_ts_step = B_ts.shape[1]
        self.num_dem_ftr = 0 if B_dem is None else B_dem.shape[1]

        # Creating all data (background and test together)
        self.all_ts = np.concatenate((self.background_ts, self.test_ts), axis=0)
        self.all_mask = None if test_mask is None else np.concatenate((self.background_mask, self.test_mask), axis=0)
        self.all_dem = None if test_dem is None else np.concatenate((self.background_dem, self.test_dem), axis=0)

        # Creating converted data for SHAP
        self.background_data = self.data_prepare(ts_x=self.background_ts, dem_x=self.background_dem, start_idx=0)
        self.test_data = self.data_prepare(ts_x=self.test_ts, dem_x=self.test_dem, start_idx=self.num_background)

    def data_prepare(self, ts_x, dem_x=None, start_idx=0):
        # Modified for sliding time window
        assert len(ts_x.shape) == 3
        assert dem_x is None or len(dem_x.shape) == 2

        total_num_features = self.num_dem_ftr + self.num_ts_ftr * self.num_window

        x_ = [[i] * total_num_features for i in range(start_idx, start_idx + ts_x.shape[0])]

        return np.array(x_)

    def wraper_predict(self, x, start_ind=0):
        assert len(x.shape) == 2

        # Calculating the indices inside the time window
        inside_ind = list(range(start_ind, start_ind + self.window_len))

        dem_x, ts_x = x[:, :self.num_dem_ftr].copy(), x[:, self.num_dem_ftr:].copy()

        # initializing the value of all arrays
        ts_x_ = np.zeros((x.shape[0], self.num_ts_step, self.num_ts_ftr))
        mask_x_ = np.zeros_like(ts_x_)
        dem_x_ = np.zeros_like(dem_x, dtype=float)
        tstep = np.ones((x.shape[0], self.num_ts_step, 1)) * \
                np.reshape(np.arange(0, self.num_ts_step), (1, self.num_ts_step, 1))

        # Reshaping the ts indices based on the num time windows and features
        ts_x = ts_x.reshape((ts_x.shape[0], self.num_window, self.num_ts_ftr))

        for i in range(x.shape[0]):
            # creating time series data
            for t in range(self.num_ts_step):
                for j in range(self.num_ts_ftr):
                    # Finding the corresponding time interval
                    wind_t = 0 if (t in inside_ind) else 1
                    ind = ts_x[i, wind_t, j]
                    ts_x_[i, t, j] = self.all_ts[ind, t, j]
                    mask_x_[i, t, j] = None if self.all_mask is None else self.all_mask[ind, t, j]
            # creating static data
            for j in range(dem_x.shape[1]):
                ind = dem_x[i, j]
                dem_x_[i, j] = None if self.all_dem is None else self.all_dem[ind, j]

        # Creating the input of the model based on the different models.
        # This part should be updated as new models get involved in the project
        if self.model_type == 'lstm_dem':
            model_input = [ts_x_, dem_x_]
        elif self.model_type == 'grud':
            model_input = [ts_x_, mask_x_, tstep]
        elif self.model_type == 'lstm':
            model_input = ts_x_
        return self.model.predict(model_input)

    def shap_values(self, num_output=1, nsamples='auto'):
        # Initializing number of time windows and contribution score matrices
        seq_len = self.background_ts.shape[1]
        num_sw = np.ceil((seq_len - self.window_len) / self.stride).astype('int') + 1
        ts_phi = np.zeros((self.num_test, num_sw, 2, self.background_ts.shape[2]))
        dem_phi = np.zeros((self.num_test, num_sw, self.num_dem_ftr))

        # Determining the number of samples
        if nsamples == 'auto':
            nsamples = 10 * self.num_ts_ftr + 5 * self.num_dem_ftr

        # Main loop on different possible windows
        for stride_cnt in range(num_sw):
            predict = lambda x: self.wraper_predict(x, start_ind=stride_cnt * self.stride)
            # Running SHAP
            self.explainer = shap.KernelExplainer(predict, self.background_data)
            shap_values = self.explainer.shap_values(self.test_data, nsamples=nsamples)
            shap_values = np.array(shap_values)

            # Extracting the SHAP values and storing them
            dem_shap_values_ = shap_values[:, :, :self.num_dem_ftr]
            ts_shap_values = shap_values[:, :, self.num_dem_ftr:]

            ts_shap_values = ts_shap_values.reshape((num_output, self.num_test, 2, self.num_ts_ftr))
            ts_phi[:, stride_cnt, :, :] = ts_shap_values[0]
            dem_phi[:, stride_cnt, :] = dem_shap_values_[0]

        # Averaging shap values from different windows -- this aggregation is done not correctly, as it does not take into account non-binary cases
        ts_phi_agg = np.empty((self.num_test, num_sw, self.num_ts_step, self.num_ts_ftr))
        ts_phi_agg[:] = np.nan
        for k in range(num_sw):
            ts_phi_agg[:, k, k * self.stride:k * self.stride + self.window_len, :] = ts_phi[:, k, 0, :][:, np.newaxis,
                                                                                     :]
        ts_phi_agg = np.nanmean(ts_phi_agg, axis=1)
        dem_phi = np.nanmean(dem_phi, axis=1)

        self.dem_phi = dem_phi
        self.ts_phi = ts_phi_agg

        return ts_phi_agg if self.num_dem_ftr == 0 else (dem_phi, ts_phi_agg)


class DynamicWindowSHAP():
    '''
    A class for computing the shapely values for time sereis data. Only the shap values for the first output
    is reported.

    Parameters:
    model: A model object that will be used for prediction. The model object must have a method called predict() which produces the model output for a given input
    delta: The treshold value in Dynamic WindowSHAP algorithm
    n_w: The maximum allowed number of time windows for each variable
    B_ts: A 3D numpy array of background time series data
    test_ts: A 3D numpy array of test time series data
    B_mask: A 3D numpy array of background masking data. It is only used for specific models such as GRUD where a masking variable is passed to the model alongside the time series data. (default: None)
    B_dem: A 2D numpy array of background demographic data (non-temporal data). It is only used for specific models with both modelities of temporal and non-temporal variables. (default: None)
    test_mask: A 3D numpy array of test mask data (default: None)
    test_dem: A 2D numpy array of test demographic data (default: None)
    model_type: The type of model being used. Set the parameter to 'lstm' when time series data is the only input, pick 'lstm_dem' when input includes both time sereis and demographic (non-termporal) data, and 'grud' when you are using GRUD structure.  (default: 'lstm')
    '''

    def __init__(self, model, delta, n_w, B_ts, test_ts, B_mask=None, B_dem=None,
                 test_mask=None, test_dem=None, model_type='lstm'):
        self.model = model
        self.model_type = model_type
        self.num_background = len(B_ts)
        self.num_test = len(test_ts)
        self.background_ts = B_ts
        self.background_mask = B_mask
        self.background_dem = B_dem
        self.test_ts = test_ts
        self.test_mask = test_mask
        self.test_dem = test_dem
        self.ts_phi = None
        self.dem_phi = None
        self.explainer = None

        # Problem sizes
        self.num_ts_ftr = B_ts.shape[2]
        self.num_ts_step = B_ts.shape[1]
        self.num_dem_ftr = 0 if B_dem is None else B_dem.shape[1]

        ## Specific to Binary Time Window
        assert self.num_test == 1  # For binary time window algorithm, samples should be fed to the algorithm one-by-one
        self.delta = delta
        self.n_w = n_w
        self.split_points = [[self.num_ts_step - 1]] * self.num_ts_ftr  # Splitting points
        self.num_window = [1] * self.num_ts_ftr

        # Creating all data (background and test together)
        self.all_ts = np.concatenate((self.background_ts, self.test_ts), axis=0)
        self.all_mask = None if test_mask is None else np.concatenate((self.background_mask, self.test_mask), axis=0)
        self.all_dem = None if test_dem is None else np.concatenate((self.background_dem, self.test_dem), axis=0)

        # Creating converted data for SHAP
        self.background_data = self.data_prepare(ts_x=self.background_ts, dem_x=self.background_dem, start_idx=0)
        self.test_data = self.data_prepare(ts_x=self.test_ts, dem_x=self.test_dem, start_idx=self.num_background)

    def data_prepare(self, ts_x, dem_x=None, start_idx=0):
        assert len(ts_x.shape) == 3
        assert dem_x is None or len(dem_x.shape) == 2
        total_num_features = self.num_dem_ftr + sum(self.num_window)  ## Specific to Binary Time Window

        x_ = [[i] * total_num_features for i in range(start_idx, start_idx + ts_x.shape[0])]

        return np.array(x_)

    def wraper_predict(self, x):
        assert len(x.shape) == 2

        dem_x, ts_x = x[:, :self.num_dem_ftr].copy(), x[:, self.num_dem_ftr:].copy()

        # initializing the value of all arrays
        ts_x_ = np.zeros((x.shape[0], self.num_ts_step, self.num_ts_ftr))
        mask_x_ = np.zeros_like(ts_x_)
        dem_x_ = np.zeros_like(dem_x, dtype=float)
        tstep = np.ones((x.shape[0], self.num_ts_step, 1)) * \
                np.reshape(np.arange(0, self.num_ts_step), (1, self.num_ts_step, 1))

        # Reshaping the ts indices based on the time windows for each feature
        ## Specific to Binary Time Window
        temp_ts_x = np.zeros((ts_x.shape[0], max(self.num_window), self.num_ts_ftr), dtype=int)
        for i in range(self.num_ts_ftr):
            temp_ts_x[:, :self.num_window[i], i] = ts_x[:, sum(self.num_window[:i]):sum(self.num_window[:i + 1])]
        ts_x = temp_ts_x

        for i in range(x.shape[0]):
            # creating time series data
            for j in range(self.num_ts_ftr):
                # Finding the corresponding time interval
                wind_t = np.searchsorted(self.split_points[j],
                                         np.arange(self.num_ts_step))  ## Specific to Binary Time Window
                for t in range(self.num_ts_step):
                    ind = ts_x[i, wind_t[t], j]
                    ts_x_[i, t, j] = self.all_ts[ind, t, j]
                    mask_x_[i, t, j] = None if self.all_mask is None else self.all_mask[ind, t, j]
            # creating static data
            for j in range(dem_x.shape[1]):
                ind = dem_x[i, j]
                dem_x_[i, j] = None if self.all_dem is None else self.all_dem[ind, j]

        # Creating the input of the model based on the different models.
        # This part should be updated as new models get involved in the project
        if self.model_type == 'lstm_dem':
            model_input = [ts_x_, dem_x_]
        elif self.model_type == 'grud':
            model_input = [ts_x_, mask_x_, tstep]
        elif self.model_type == 'lstm':
            model_input = ts_x_

        return self.model.predict(model_input)

    def shap_values(self, num_output=1, nsamples_in_loop='auto', nsamples_final='auto'):
        flag = 1
        while flag:
            flag = 0
            # Updating the number of time windows for each time series feature
            self.num_window = [len(self.split_points[i]) for i in range(self.num_ts_ftr)]

            # Updating converted data for SHAP
            self.background_data = self.data_prepare(ts_x=self.background_ts, dem_x=self.background_dem, start_idx=0)
            self.test_data = self.data_prepare(ts_x=self.test_ts, dem_x=self.test_dem, start_idx=self.num_background)

            # Running SHAP
            if nsamples_in_loop == 'auto':
                nsamples = 2 * sum(self.num_window)
            else:
                nsamples = nsamples_in_loop

            self.explainer = shap.KernelExplainer(self.wraper_predict, self.background_data)
            shap_values = self.explainer.shap_values(self.test_data, nsamples=nsamples)
            shap_values = np.array(shap_values)
            dem_phi = shap_values[0, :, :self.num_dem_ftr]  # Extracting dem SHAP values
            ts_shap_values = shap_values[:, :, self.num_dem_ftr:]  # Extracting ts SHAP values

            # Checking the maximum number of windows condition
            if max(self.num_window) >= self.n_w: break

            for i in range(self.num_ts_ftr):
                S = set(self.split_points[i])
                for j in range(self.num_window[i]):
                    if abs(ts_shap_values[0, 0, sum(self.num_window[:i]) + j]) > self.delta:
                        S.add(int(self.split_points[i][j] / 2) if j == 0 else int(
                            (self.split_points[i][j - 1] + self.split_points[i][j]) / 2))
                if set(S) != set(self.split_points[i]):
                    flag += 1
                    self.split_points[i] = list(S)
                    self.split_points[i].sort()

        # Running SHAP with large number of samples for the final evaluation of Shapely values
        self.explainer = shap.KernelExplainer(self.wraper_predict, self.background_data)
        shap_values = self.explainer.shap_values(self.test_data, nsamples=nsamples_final)
        shap_values = np.array(shap_values)
        dem_phi = shap_values[0, :, :self.num_dem_ftr]  # Extracting dem SHAP values
        ts_shap_values = shap_values[:, :, self.num_dem_ftr:]  # Extracting ts SHAP values

        # Assigning Shap values to each single time step
        ts_phi = np.zeros((self.num_test, self.num_ts_step, self.num_ts_ftr))
        for i in range(self.num_ts_ftr):
            for j in range(self.num_window[i]):
                # This part of the code is written in a way that each splitting point belongs to the time window that starts from that point
                # For the last time window, both splitting points at the end and start of the time window belong to it
                start_ind = 0 if j == 0 else self.split_points[i][j - 1]
                end_ind = self.split_points[i][j] + int((j + 1) / self.num_window[i])
                ts_phi[0, start_ind:end_ind, i] = ts_shap_values[0, :, sum(self.num_window[:i]) + j] / (
                            end_ind - start_ind)
        self.dem_phi = dem_phi
        self.ts_phi = ts_phi

        return ts_phi if self.num_dem_ftr == 0 else (dem_phi, ts_phi)

